#include "headers.h"
#include "Tablet.h"

#define LOGNAME	"tablet"
#include "diag.h"

static const int transfer_rate_values[8] = 
{
	1, 5, 11, 23, 47, 95, 95, 95
};


Tablet::Tablet()
: Device(DEV_ID('8', '8', '7', '5')), access(0), scheduler(0),
  wintab_enabled(false), hctx(NULL), hthread(NULL), hwnd(NULL),
  tabletname(NULL), checking(false), setting(false)
{
	InitializeCriticalSection(&cs);
	config.mode = MODE_DEFAULT;
}

Tablet::~Tablet()
{
	DeleteCriticalSection(&cs);
}

bool Tablet::Init(IIOAccess* _access, IScheduler* _scheduler, ITime* _tim, 
				  HINSTANCE hinst, uint _prxrdy, uint _preq)
{
	access		= _access;
	scheduler	= _scheduler;
	tim			= _tim;
	prxrdy		= _prxrdy;
	prequest	=_preq;
	port30		= 0;

	hwnd = CreateWindow("STATIC", "PC-8875", WS_OVERLAPPEDWINDOW,
						CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hinst, NULL);

	if (!LoadWintab())
	{
		strcpy(message, "Wintab not available.");
		return true;
	}
	else if (!gpWTInfoA(0, 0, NULL))
	{
		strcpy(message, "WinTab Services Not Available.");
		UnloadWintab();
		return true;
	}
	else
	{
		// Get device name
		UINT len = gpWTInfoA(WTI_DEVICES, DVC_NAME, NULL);
		tabletname = new char[len + 1];
		if (len > 0)
			len = gpWTInfoA(WTI_DEVICES, DVC_NAME, tabletname);
		tabletname[len] = 0;
		wintab_enabled = true;
	}

	Close();
	Reset();
	return true;
}

void Tablet::Release()
{
	Close();
	if (hwnd != NULL)
		DestroyWindow(hwnd);
	if (tabletname)
		delete[] tabletname;
	UnloadWintab();
}

// ---------------------------------------------------------------------------
//	I/O |[gĎ

void IFCALL Tablet::Reset(uint, uint)
{
	datacount		= 0;
	datapos			= 0;
	statuscount		= 0;
	mode			= Mode::clear;
	status.Reset();
	pc8875.Reset();
	mypad.Reset();
	creator.Reset();
	mt20.Reset();
	cat.Reset();
}

void IFCALL Tablet::Out20(uint, uint data)
{
	if (IsRs232cEnabled())
	{
		byte buf[256];
		uint len = 0;
		switch(config.mode)
		{
		case TabletMode::ascii_binary:
		case TabletMode::bcd:
			len = pc8875.Out20(data, buf);
			if (len > 0)
			{
				checking = true;
				SetData(buf, len);
				checking = false;
			}
			break;
		case TabletMode::logitec:
			mypad.Out20(data);
			break;
		}
		statuscount = 0;
	}
}

void IFCALL Tablet::Out21(uint, uint d)
{
	//LOG1("[%.2x] ", d);
	
	switch (mode)
	{
	case clear:
		// Mode Instruction
		if (d & 3)
		{
			// Asynchronus mode
			mode = async;
		}
		else
		{
			// Synchronus mode
			mode = sync1;
		}
		break;
		
	case sync1:
		mode = sync2;
		break;

	case sync2:
		mode = sync;
		LOG0("\n");
		break;

	case async:
	case sync:
		// Command Instruction
		// b7 - enter hunt mode
		// b6 - internal reset
		if (d & 0x40)
		{
			// Reset!
			LOG0(" Internal Reset!\n");
			Reset();
			mode = clear;
			access->In(0x20);	// clear RXRDY
			break;
		}
		break;
	default:
		LOG1("internal error? <%d>\n", mode);
		break;
	}
}

void IFCALL Tablet::Out30(uint, uint data)
{
	// BIT 5,4	00...CMT 600baud
	//			01...CMT 1200baud
	//			10...RS232C ASYNC
	//			11...RS232C SYNC
	port30 = data;
}

void IFCALL Tablet::Out40(uint, uint data)
{
	if (IsTabletEnabled() && config.mode == TabletMode::busmouse)
	{
		neos.Out40(data, tim->GetTime());
	}
}

void IFCALL Tablet::Out44(uint, uint data)
{
	if (IsTabletEnabled() && config.mode == TabletMode::busmouse)
	{
		neos.Out44(data);
	}
}

void IFCALL Tablet::Out80(uint, uint data)
{
	if (IsTabletEnabled() && config.mode == TabletMode::wacom)
	{
		creator.Out80(data);
	}
}

uint IFCALL Tablet::In0x(uint port)
{
	uint ret = 0xff;
	if (hctx != NULL)
	{
		if (status.keypress1 && config.usekey1 && config.keyport1 == port)
			ret ^= (1 << config.keybit1);
		if (status.keypress2 && config.usekey2 && config.keyport2 == port)
			ret ^= (1 << config.keybit2);
	}
	return ret;
}

uint IFCALL Tablet::In20(uint)
{
	if (IsRs232cEnabled() && mode != clear)
	{
		//LOG0("In20\n");
		int count = 150;
		scheduler->DelEvent(this);
		scheduler->AddEvent(count, this, STATIC_CAST(TimeFunc, &Tablet::TimeEvent), 1);
		statuscount = 0;
	}
	return 0xff;
}

uint IFCALL Tablet::In21(uint)
{
	if (!checking && !setting) 
	{
		checking = true;
	
		if (IsRs232cEnabled())
		{
			//LOG0("In21\n");
			statuscount++;
			if (statuscount > 4)
			{
				// Xe[^X|[OAvdȂւ̑΍
				SetData(NULL, 0);
				statuscount = 0;
			}
		}
		checking = false;
	}
	return 0xff;
}

uint IFCALL Tablet::In45(uint)
{
	return (IsTabletEnabled() && config.mode == TabletMode::busmouse) ?
		   neos.In45() : 0xff;
}

uint IFCALL Tablet::In80(uint)
{
	return (IsTabletEnabled() && config.mode == TabletMode::wacom) ?
		   creator.In80() : 0xff;
}

uint IFCALL Tablet::In81(uint)
{
	return (IsTabletEnabled() && config.mode == TabletMode::wacom) ?
		   creator.In81() : 0xff;
}

uint IFCALL Tablet::InBx(uint port)
{
	return (IsTabletEnabled() && config.mode == TabletMode::hal) ?
		   cat.InBx(port) : 0xff;
}

// ---------------------------------------------------------------------------
//	Tablet
//

bool Tablet::Open(TABLET_CONFIG* pconfig)
{
	if (!wintab_enabled) 
		return false;
	if (pconfig->mode != TabletMode::disabled && hctx != NULL)
	{
		config = *pconfig;
		ApplyConfig();
		return true;
	}

	Close();
	config = *pconfig;
	if (config.mode == TabletMode::disabled)
		return true;

	// Get default context information
	gpWTInfoA( WTI_DEFCONTEXT, 0, &lc );

	// Open the context
	lc.lcPktData = PACKETDATA;
	lc.lcPktMode = PACKETMODE;
	hctx = gpWTOpenA(hwnd, &lc, TRUE );
	if (NULL == hctx)
	{
		strcpy(message, "Tablet device open error.");
		return false;
	}

	ApplyConfig();

	if (!hthread)
	{
		running = true;
		hthread = (HANDLE)(_beginthreadex(NULL, 0, ThreadEntry, 
						   reinterpret_cast<void*> (this), 0, &idthread));
	}
	if (!hthread)
	{
		Close();
		strcpy(message, "Thread setup error.");
		return false;
	}

	return true;
}

void Tablet::Close()
{
	if (!wintab_enabled) 
		return;

	pc8875.Close();

	if (NULL != hctx)
	{
		gpWTClose(hctx);
	}

	DWORD x = GetTickCount();

	if (hthread)
	{
		running = false;
		if (WAIT_TIMEOUT == WaitForSingleObject(hthread, 1000))
		{
			TerminateThread(hthread, 0);
		}
		CloseHandle(hthread);
	}

	Reset();
	hctx = NULL;
	hthread = NULL;
	strcpy(message, "Disabled");
}

bool Tablet::GetPacket(PACKET& packet)
{
	const int packet_num = 16;
	PACKET pkt[packet_num];

	int total = 0;
	int num = 0;
	do 
	{
		num = gpWTPacketsGet(hctx, packet_num, pkt );
		if (num > 0)
		{
			packet = pkt[num - 1];
			total += num;
		}
	} while (num > 0);

	//if (total > 0)
	//{
	//	LOG4("%d : %d, %d, %d\n", total, packet.pkX, packet.pkY, packet.pkButtons);
	//}
	return (total > 0);
}

void Tablet::ApplyConfig()
{
	switch(config.mode)
	{
	case TabletMode::ascii_binary:
		pc8875.Open(Pc8875::Format::ascii_binary, lc);
		break;
	case TabletMode::bcd:
		pc8875.Open(Pc8875::Format::bcd, lc);
		break;
	case TabletMode::logitec:
		mypad.Open(lc);
		break;
	case TabletMode::wacom:
		creator.Open(lc);
		break;
	case TabletMode::epson:
	case TabletMode::mouse:
		mt20.Open(config.mouse_sens, lc);
		break;
	case TabletMode::hal:
		cat.Open(config.mouse_sens, lc);
		break;
	case TabletMode::busmouse:
		neos.Open(config.mouse_sens, lc);
		break;
	}
	Reset();
}

char* Tablet::GetMessage()
{
	if (IsTabletEnabled())
	{
		sprintf(message, "Enabled\n%s\nResolution : %dx%d\n", 
			tabletname ? tabletname : "", lc.lcInExtX, lc.lcInExtY);
		switch(config.mode)
		{
		case TabletMode::ascii_binary:
		case TabletMode::bcd:
			pc8875.SetStatusMessage(message);
			break;
		case TabletMode::logitec:
			mypad.SetStatusMessage(message);
			break;
		}
	}
	return message;
}

// ---------------------------------------------------------------------------
//	RS-232C
//

void Tablet::SetData(byte* data, uint count)
{
	EnterCriticalSection(&cs);

	for (uint i=0; i<count; i++)
	{
		databuf[(datapos + datacount) % DATABUF_SIZE] = data[i];
		datacount++;
		//LOG4("%c %02X,%d,%d\n", (isprint(data[i]) ? data[i] : '.'),  
		//	data[i], datapos, datacount);
	}

	uint in21 = access->In(0x21);
	if (datacount > 0)
	{
		DWORD t = timeGetTime();
		if ((in21 & 0x02) == 0)
		{
			//char bbb[256];
			//sprintf(bbb,"R:%02X %d %d\r\n",databuf[datapos], datapos, datacount);
			//OutputDebugString(bbb);
			//LOG4("\t%c %02X,%d,%d\n", (isprint(databuf[datapos]) ? databuf[datapos] : '.'),  
			//	databuf[datapos], datapos, datacount);
			access->Out(prequest, databuf[datapos]);
			datapos = (datapos + 1) % DATABUF_SIZE;
			datacount--;
			access->Out(prxrdy, 1);
			tick_setdata = t;
		}
		else if ((t - tick_setdata) > (DWORD)((130 >> 7) + 5))
		{
			access->Out(prxrdy, 1);
			tick_setdata = t;
		}
	}

	LeaveCriticalSection(&cs);
}

// ---------------------------------------------------------------------------
//	thread
//
uint Tablet::ThreadMain()
{
	LOG0("Thread start\n");
	byte buf[256];

	char	bb[256];
	DWORD	x = 0;
	DWORD	x1 = 0;
	DWORD	x2 = 0;
	PACKET	packet;
	int		buf_max;

	memset(&packet, 0, sizeof(PACKET));
	timeBeginPeriod(1);

	while (running)
	{
		bool get_packet = GetPacket(packet);
		x2 = timeGetTime();
		bool valid = status.SetValid(get_packet, x2, packet);

		DWORD len = 0;
		buf[0] = 0;
		switch(config.mode)
		{
		case TabletMode::ascii_binary:
		case TabletMode::bcd:
			buf_max = 10 * (config.mode == TabletMode::ascii_binary ? 9 : 12);
			if (datacount < buf_max)
			{
				len = pc8875.SetValue(buf, packet, valid, x2);
			}
			break;
		case TabletMode::logitec:
			if (datacount < 50)
			{
				len = mypad.SetValue(buf, packet, valid, x2);
			}
			break;
		case TabletMode::wacom:
			creator.SetValue(packet, valid);
			break;
		case TabletMode::epson:
			len = mt20.SetValueTablet(buf, packet, valid, x2);
			break;
		case TabletMode::mouse:
			len = mt20.SetValueMouse(buf, packet, valid, x2);
			break;
		case TabletMode::hal:
			cat.SetValue(packet, valid);
			break;
		case TabletMode::busmouse:
			neos.SetValue(packet, valid);
			break;
		}

		setting = true;
		SetData(buf, len);
		setting = false;

		x1 = timeGetTime();
		Sleep(1);
		DWORD y = timeGetTime();
		sprintf(bb,"%d %d %d %d %d %s\r\n", x, y-x, y-x2, y-x1, len, buf);
		//OutputDebugString(bb);
		x = y;
	}
	LOG0("Thread end\n");
	timeEndPeriod(1);
	return 0;
}

uint __stdcall Tablet::ThreadEntry(LPVOID arg)
{
	if (arg)
		return reinterpret_cast<Tablet*> (arg)->ThreadMain();
	else
		return 0;
}

// ---------------------------------------------------------------------------
//	Timer event
//
void IOCALL Tablet::TimeEvent(uint e)
{
	SetData(NULL, 0);
}

// ---------------------------------------------------------------------------
//	device description
//
const Device::Descriptor Tablet::descriptor = { indef, outdef };

const Device::OutFuncPtr Tablet::outdef[] = 
{
	STATIC_CAST(Device::OutFuncPtr, Reset),
	STATIC_CAST(Device::OutFuncPtr, Out20),
	STATIC_CAST(Device::OutFuncPtr, Out21),
	STATIC_CAST(Device::OutFuncPtr, Out30),
	STATIC_CAST(Device::OutFuncPtr, Out40),
	STATIC_CAST(Device::OutFuncPtr, Out44),
	STATIC_CAST(Device::OutFuncPtr, Out80),
};

const Device::InFuncPtr Tablet::indef[] = 
{
	STATIC_CAST(Device::InFuncPtr, In0x),
	STATIC_CAST(Device::InFuncPtr, In20),
	STATIC_CAST(Device::InFuncPtr, In21),
	STATIC_CAST(Device::InFuncPtr, In45),
	STATIC_CAST(Device::InFuncPtr, In80),
	STATIC_CAST(Device::InFuncPtr, In81),
	STATIC_CAST(Device::InFuncPtr, InBx),
};
