#include "headers.h"
#include "pc8875.h"

#define LOGNAME	"pc8875"
#include "diag.h"

static const int transfer_rate_values[8] = 
{
	1, 5, 11, 23, 47, 95, 95, 95
};


Pc8875::Pc8875()
{
}

Pc8875::~Pc8875()
{
}

bool Pc8875::Init()
{
	Close();
	Reset();
	return true;
}

void Pc8875::Release()
{
	Close();
}

// ---------------------------------------------------------------------------
//	I/O |[gĎ

void Pc8875::Reset()
{
	readmode		= ReadMode::point;
	transfermode	= TransferMode::single;
	transferrate	= 95;
	command			= 0;
	last_pen		= false;
	last_valid		= false;
	point_pen		= false;
	readstatus		= ReadStatus::stop;
}

uint Pc8875::Out20(uint data, byte* buf)
{
	uint len = 0;

	switch(format)
	{
	case Format::ascii_binary:
		switch(command) 
		{
		case 0:
			if (data == 0x1b) command++;
			break;
		case 1:
			command = (data == '!' ? 2 : 0);
			break;
		case 2:
			len = SetAsciiBinaryCommand(data, buf);
			command = 0;
			break;
		default:
			command = 0;
			break;
		}
		break;
	case Format::bcd:
		len = SetBcdCommand(data, buf);
		break;
	}
	return len;
}

// ---------------------------------------------------------------------------
//	RS-232C
//

bool Pc8875::Open(Format _format, LOGCONTEXT _lc)
{
	Reset();
	format = _format;
	lc = _lc;
	return true;
}

void Pc8875::Close()
{
}

void Pc8875::SetStatusMessage(char *message)
{
	char buf[256];
	static const LPSTR transfermodename[] = 
	{ 
		"Single point mode", "Multiple point mode" 
	};
	static const LPSTR readmodename[] = 
	{ 
		"Point mode", "Switch stream mode", "Presence stream mode", "Increment mode" 
	};
	sprintf(buf, "Read mode : %s\nTransfer mode : %s\nTransfer rate : %d point/sec\n", 
		readmodename[readmode & 3], transfermodename[transfermode & 1], transferrate);
	strcat(message, buf);
}

uint Pc8875::SetAsciiBinaryCommand(uint data, byte* buf)
{
	if (data < 0x20) 
		return 0;
	data &= 0x0f;

	uint len = 0;
	if (data < 8)
	{
		switch(data & 2)
		{
		case 0:
			ChangeTransferMode(TransferMode::single);
			break;
		case 2:
			ChangeTransferMode(TransferMode::multiple);
			break;
		}
		switch(data & 0x05)
		{
		case 0:
			len = ChangeReadMode(ReadMode::switch_stream, buf);
			break;
		case 1:
			len = ChangeReadMode(ReadMode::presence_stream, buf);
			break;
		case 4:
			len = ChangeReadMode(ReadMode::point, buf);
			break;
		case 5:
			len = ChangeReadMode(ReadMode::increment, buf);
			break;
		}
	}
	else if (data < 10)
	{
		readstatus = ReadStatus::stop;
	}
	else 
	{
		transferrate = transfer_rate_values[data - 10];
	}
	return len;
}

uint Pc8875::SetBcdCommand(uint data, byte* buf)
{
	uint len = 0;

	switch(data)
	{
	case 'Q':
		ChangeTransferMode(TransferMode::single);
		break;
	case 'R':
		ChangeTransferMode(TransferMode::multiple);
		break;
	case 'P':
		len = ChangeReadMode(ReadMode::point, buf);
		break;
	}
	SetBcdCommandSub(ReadMode::switch_stream, data, '@', len, buf);
	SetBcdCommandSub(ReadMode::presence_stream, data, 'H', len, buf);
	SetBcdCommandSub(ReadMode::increment, data, 'S', len, buf);

	return len;
}

void Pc8875::SetBcdCommandSub(ReadMode rm, uint data, uint top, uint& len, byte* buf)
{
	if (data >= top && data < top + 8)
	{
		len = ChangeReadMode(rm, buf);
		transferrate = transfer_rate_values[data - top];
	}
}

void Pc8875::ChangeTransferMode(TransferMode tm)
{
	LOG2("TransferMode %d -> %d\r\n", transfermode, tm);
	transfermode = tm;
}

uint Pc8875::ChangeReadMode(ReadMode rm, byte* buf)
{
	LOG2("ReadMode %d -> %d\r\n", readmode, rm);

	uint len = 0;
	if (readmode == rm)
	{
		if (transfermode == TransferMode::single)
		{
			switch(readstatus)
			{
			case ReadStatus::swoff:
				len = SetValueString(buf, send_x, send_y, false, true);
				readstatus = invalid_oarea ? ReadStatus::oarea : ReadStatus::stop;
				break;
			case ReadStatus::oarea:
				len = SetValueString(buf, send_x, send_y, false, false);
				readstatus = ReadStatus::stop;
				point_pen = false;
				break;
			default:
				readstatus = ReadStatus::collect;
				break;
			}
		}
		else
		{
			last_valid = false;
			readstatus = ReadStatus::collect;
		}
	}
	else
	{
		readmode = rm;
		readstatus = ReadStatus::collect;
	}

	return len;
}

uint Pc8875::SetValue(byte* buf, PACKET& packet, bool valid, DWORD tick)
{
	const int max_x = 3071;
	const int max_y = 2047;

	uint	len = 0;
	bool	set_value = false;
	int x = Limit((max_x * packet.pkX) / lc.lcInExtX, max_x, 0);
	int y = Limit((max_y * packet.pkY) / lc.lcInExtY, max_y, 0);
	bool pen = ((packet.pkButtons & 1) != 0);
	switch(readmode)
	{
	case ReadMode::point:
		set_value = SetPointValue(x, y, pen, valid);
		break;
	case ReadMode::switch_stream:
		set_value = SetSwitchStreamValue(x, y, pen, valid, tick);
		break;
	case ReadMode::presence_stream:
	case ReadMode::increment:
		set_value = SetPresenceStreamValue(x, y, pen, valid, tick);
		break;
	}
	if (set_value)
	{
		len = SetValueString(buf, x, y, pen, valid);
	}
	last_x = x;
	last_y = y;
	last_pen = pen;
	last_valid = valid;

	return len;
}

uint Pc8875::SetValueString(byte* buf, int x, int y, bool pen, bool valid)
{
	uint len = 0;
	int header = last_valid ? 0x1a : 0x1b;
	if (!valid)
	{
		x = send_x;
		y = send_y;
	}
	int pen_value = pen ? 1 : 0;

	switch(format)
	{
	case Format::ascii_binary:
		buf[0] = valid ? header : 0x1f;
		buf[1] = 0x20 | pen_value;
		buf[2] = 0x20 | ((y << 1) & 0x1e);
		buf[3] = 0x20 | ((y >> 4) & 0x1f);
		buf[4] = 0x20 | ((y >> 9) & 0x07);
		buf[5] = 0x20 | ((x << 1) & 0x1e);
		buf[6] = 0x20 | ((x >> 4) & 0x1f);
		buf[7] = 0x20 | ((x >> 9) & 0x07);
		buf[8] = 0x0d;
		buf[9] = 0;
		len = 9;
		break;
	case Format::bcd:
		sprintf((char*)buf, "%04d,%04d,%d\r", x, y, valid ? pen_value : 8);
		len = 12;
		break;
	default:
		break;
	}
	//if (len > 0)
	//{
	//	char bbb[256];
	//	sprintf(bbb,"%d %d %d %d %d %s\r\n", x, y, pen, valid, datacount, buf);
	//	OutputDebugString(bbb);
	//}
	return len;
}

bool Pc8875::SetPointValue(int x, int y, bool pen, bool valid)
{
	bool ret = false;
	if (valid && !last_pen && pen && readstatus == ReadStatus::collect)
	{
		send_x		= x;
		send_y		= y;
		ret			= true;
		point_pen	= true;
		if (transfermode == TransferMode::single)
		{
			readstatus = ReadStatus::wait;
		}
	}
	if (last_valid && !valid && point_pen)
	{
		if (transfermode == TransferMode::multiple || readstatus == ReadStatus::collect)
		{
			ret = true;
		}
		else if (transfermode == TransferMode::single && readstatus == ReadStatus::wait)
		{
			readstatus = ReadStatus::oarea;
		}
		point_pen = false;
	}
	return ret;
}

bool Pc8875::SetSwitchStreamValue(int x, int y, bool pen, bool valid, DWORD tick)
{
	bool ret = false;
	if (valid && pen && readstatus == ReadStatus::collect)
	{
		DWORD diff = (1000 / transferrate);
		if (!last_pen || (tick - last_tick) > diff)
		{
			send_x		= x;
			send_y		= y;
			ret			= true;
			point_pen	= true;
			last_tick = tick;
			if (transfermode == TransferMode::single)
			{
				readstatus = ReadStatus::wait;
			}
		}
	}
	if (last_pen && !pen)
	{
		if (transfermode == TransferMode::multiple || readstatus == ReadStatus::collect)
		{
			send_x	= x;
			send_y	= y;
			ret		= true;
		} 
		else if (transfermode == TransferMode::single && ReadStatus::wait)
		{
			send_x			= x;
			send_y			= y;
			invalid_oarea	= false;
			readstatus		= ReadStatus::swoff;
		}
	}
	if (last_valid && !valid && point_pen)
	{
		if (transfermode == TransferMode::multiple || readstatus == ReadStatus::collect)
		{
			ret = true;
		}
		else if (transfermode == TransferMode::single)
		{
			switch(readstatus)
			{
			case ReadStatus::wait:
				readstatus = ReadStatus::oarea;
				break;
			case ReadStatus::swoff:
				invalid_oarea = true;
				break;
			}
		}
		point_pen = false;
	}
	return ret;
}

bool Pc8875::SetPresenceStreamValue(int x, int y, bool pen, bool valid, DWORD tick)
{
	bool ret = false;
	if (valid && readstatus == ReadStatus::collect)
	{
		if (readmode == ReadMode::presence_stream || 
			send_x != x || send_y != y || last_pen != pen)
		{
			DWORD diff = (1000 / transferrate);
			if (last_pen != pen || (tick - last_tick) > diff)
			{
				send_x		= x;
				send_y		= y;
				send_pen	= pen;
				ret			= true;
				point_pen	= true;
				last_tick	= tick;
				if (transfermode == TransferMode::single)
				{
					readstatus = ReadStatus::wait;
				}
			}
		}
	}
	if (last_valid && !valid && point_pen)
	{
		if (transfermode == TransferMode::multiple || readstatus == ReadStatus::collect)
		{
			ret = true;
		}
		else if (transfermode == TransferMode::single && readstatus == ReadStatus::wait)
		{
			if (send_pen)
			{
				send_x		= x;
				send_y		= y;
				readstatus	= ReadStatus::swoff;
			}
			else
			{
				readstatus	= ReadStatus::oarea;
			}
		}
		point_pen = false;
	}
	return ret;
}
